from typing import List, Optional

from textual.app import App, ComposeResult
from textual.widgets import Header, Tree

from manifestoo_core.addon import Addon
from manifestoo_core.addons_set import AddonsSet
from manifestoo_core.core_addons import (
    is_core_addon,
    is_core_ce_addon,
    is_core_ee_addon,
)
from manifestoo_core.odoo_series import OdooEdition, OdooSeries

from ..addons_selection import AddonsSelection


class Context:
    def __init__(
        self,
        addons_set: AddonsSet,
        odoo_series: OdooSeries,
        fold_core_addons: bool,
    ) -> None:
        self.addons_set = addons_set
        self.odoo_series = odoo_series
        self.fold_core_addons = fold_core_addons


class Node:
    def __init__(
        self,
        addon_name: str,
        addon: Optional[Addon],
        context: Context,
    ):
        self.addon_name = addon_name
        self.addon = addon
        self.context = context

    def __str__(self) -> str:
        return f"{self.addon_name} {self._version_str}"

    @property
    def _version_str(self) -> str:
        if not self.addon:
            return "✘ not installed"
        elif is_core_ce_addon(self.addon_name, self.context.odoo_series):
            return f"{self.context.odoo_series.value}+{OdooEdition.CE.value}"
        elif is_core_ee_addon(self.addon_name, self.context.odoo_series):
            return f"{self.context.odoo_series.value}+{OdooEdition.EE.value}"
        else:
            return self.addon.manifest.version or "no version"

    def has_children(self) -> bool:
        return bool(self.children())

    def children(self) -> List["Node"]:
        if not self.addon:
            return []
        if self.context.fold_core_addons and is_core_addon(
            self.addon_name, self.context.odoo_series
        ):
            return []
        children = []
        for depend in self.addon.manifest.depends:
            if depend == "base":
                continue
            children.append(
                Node(
                    depend,
                    self.context.addons_set.get(depend),
                    self.context,
                )
            )
        return children


class ManifestooTree(Tree[Node]):
    def __init__(self, root_nodes: List[Node]) -> None:
        super().__init__(label="...")
        for root_node in root_nodes:
            self.root.add(
                str(root_node),
                data=root_node,
                expand=False,
                allow_expand=root_node.has_children(),
            )

    def on_tree_node_expanded(self, event: Tree.NodeExpanded[Node]) -> None:
        if event.node._children:
            # aldready expanded once
            return
        if event.node.data is None:
            # to please mypy
            return
        for child in event.node.data.children():
            event.node.add(
                str(child),
                data=child,
                expand=False,
                allow_expand=child.has_children(),
            )


class ManifestooTreeApp(App[None]):
    """Manifestoo tree."""

    def __init__(self, root_nodes: List[Node]) -> None:
        super().__init__()
        self.root_nodes = root_nodes

    def compose(self) -> ComposeResult:
        yield Header()
        tree = ManifestooTree(self.root_nodes)
        tree.show_root = False
        yield tree


def interactive_tree_command(
    addons_selection: AddonsSelection,
    addons_set: AddonsSet,
    odoo_series: OdooSeries,
    fold_core_addons: bool,
) -> None:
    context = Context(addons_set, odoo_series, fold_core_addons)
    root_nodes = [
        Node(
            addon_name,
            addons_set.get(addon_name),
            context,
        )
        for addon_name in addons_selection
    ]
    app = ManifestooTreeApp(root_nodes)
    app.title = "Manifestoo tree"
    app.run()



import sys
from typing import Dict, Iterable, Set

import typer

from manifestoo_core.addons_set import AddonsSet

from . import echo
from .exceptions import CycleErrorExit


class AddonSorter:
    @staticmethod
    def from_name(name: str) -> "AddonSorter":
        if name == "alphabetical":
            return AddonSorterAlphabetical()
        elif name == "topological":
            if sys.version_info < (3, 9):
                echo.error(
                    "The 'topological' sorter requires Python 3.9 or later",
                    err=False,
                )
                raise typer.Exit(1)
            return AddonSorterTopological()
        else:
            echo.error(f"Unknown sorter {name}", err=False)
            raise typer.Exit(1)

    def sort(
        self, addons_selection: Iterable[str], addon_set: AddonsSet
    ) -> Iterable[str]:
        raise NotImplementedError()


class AddonSorterAlphabetical(AddonSorter):
    def sort(
        self, addons_selection: Iterable[str], addon_set: AddonsSet
    ) -> Iterable[str]:
        return sorted(addons_selection)


class AddonSorterTopological(AddonSorter):
    def sort(
        self, addons_selection: Iterable[str], addon_set: AddonsSet
    ) -> Iterable[str]:
        result_dict: Dict[str, Set[str]] = {}
        for addon_name in addons_selection:
            try:
                addon = addon_set[addon_name]
                result_dict[addon_name] = set(
                    depend
                    for depend in addon.manifest.depends
                    if depend in addons_selection
                )
            except KeyError:
                echo.debug(f"Addon {addon_name} not found in addon set")
        from graphlib import CycleError, TopologicalSorter

        topological_sorted_res = TopologicalSorter(result_dict)
        try:
            res = list(topological_sorted_res.static_order())
        except CycleError as e:
            echo.error("Cycle detected in dependencies", err=False)
            raise CycleErrorExit(1) from e
        return res



import sys
from pathlib import Path
from typing import Any, Dict

from packaging.version import Version
from typer.testing import CliRunner as TyperCliRunner

from manifestoo.addons_selection import AddonsSelection
from manifestoo_core.addon import Addon
from manifestoo_core.addons_set import AddonsSet
from manifestoo_core.manifest import Manifest

if sys.version_info < (3, 8):
    from importlib_metadata import version  # type: ignore[import]
else:
    from importlib.metadata import version


def populate_addons_dir(addons_dir: Path, addons: Dict[str, Dict[str, Any]]):
    if not addons_dir.is_dir():
        addons_dir.mkdir()
    for addon_name, manifest in addons.items():
        addon_path = addons_dir / addon_name
        addon_path.mkdir()
        (addon_path / "__init__.py").touch()
        manifest_path = addon_path / "__manifest__.py"
        manifest_path.write_text(repr(manifest))


def mock_addons_set(addons: Dict[str, Dict[str, Any]]) -> AddonsSet:
    addons_set = AddonsSet()
    for addon_name, manifest_dict in addons.items():
        manifest = Manifest.from_dict(manifest_dict)
        manifest_path = Path("/tmp/fake-addons-dir") / addon_name / "__manifest__.py"
        addons_set[addon_name] = Addon(manifest, manifest_path)
    return addons_set


def mock_addons_selection(addon_names: str) -> AddonsSelection:
    addons_selection = AddonsSelection()
    addons_selection.add_addon_names(addon_names)
    return addons_selection


def CliRunner():
    if Version(version("click")) < Version("8.2"):
        # Remove this with dropping Python 3.9 support because click 8.2 is
        # python 3.10+ only
        return TyperCliRunner(mix_stderr=False)  # Avoid mixing stderr with stdout
    return TyperCliRunner()



from typing import Iterable, Optional

from manifestoo_core.addons_set import AddonsSet

from ..addon_sorter import AddonSorter, AddonSorterAlphabetical
from ..addons_selection import AddonsSelection


def list_command(
    addons_selection: AddonsSelection,
    addons_set: AddonsSet,
    addon_sorter: Optional[AddonSorter] = None,
) -> Iterable[str]:
    if not addon_sorter:
        addon_sorter = AddonSorterAlphabetical()
    return addon_sorter.sort(addons_selection, addons_set)



from textwrap import dedent

from manifestoo.commands.list import list_command
from manifestoo.main import app

from .common import (
    CliRunner,
    mock_addons_selection,
    mock_addons_set,
    populate_addons_dir,
)


def test_basic():
    addons_selection = mock_addons_selection("b,a")
    addons_set = mock_addons_set(
        {
            "a": {},
            "b": {},
        }
    )
    assert list_command(addons_selection, addons_set) == ["a", "b"]


def test_integration(tmp_path):
    addons = {
        "a": {},
        "b": {},
    }
    populate_addons_dir(tmp_path, addons)
    runner = CliRunner()
    result = runner.invoke(
        app,
        [f"--select-addons-dir={tmp_path}", "list"],
        catch_exceptions=False,
    )
    assert not result.exception
    assert result.exit_code == 0, result.stderr
    assert result.stdout == "a\nb\n"


def test_found(tmp_path):
    addons = {"a": {}, "b": {}, "c": {"installable": False}}
    populate_addons_dir(tmp_path, addons)
    conf = tmp_path / "odoo.conf"
    conf.write_text(
        dedent(
            f"""\
            [options]
            addons_path = {tmp_path}
            """
        )
    )
    runner = CliRunner()
    result = runner.invoke(
        app,
        ["--select-found", "list"],
        catch_exceptions=False,
        env={"ODOO_RC": str(conf)},
    )
    assert not result.exception
    assert result.exit_code == 0, result.stderr
    assert result.stdout == "a\nb\n"



import sys

if sys.version_info >= (3, 8):
    import importlib.metadata as importlib_metadata
else:
    import importlib_metadata

version = importlib_metadata.version("manifestoo")
core_version = importlib_metadata.version("manifestoo_core")



from typing import Iterable, Optional, Set, Tuple

from manifestoo_core.addons_set import AddonsSet

from ..addon_sorter import AddonSorter, AddonSorterAlphabetical
from ..addons_selection import AddonsSelection
from ..dependency_iterator import dependency_iterator


def list_depends_command(
    addons_selection: AddonsSelection,
    addons_set: AddonsSet,
    transitive: bool = False,
    include_selected: bool = False,
    addon_sorter: Optional[AddonSorter] = None,
) -> Tuple[Iterable[str], Iterable[str]]:
    if not addon_sorter:
        addon_sorter = AddonSorterAlphabetical()
    result: Set[str] = set()
    missing: Set[str] = set()
    for addon_name, addon in dependency_iterator(
        addons_selection, addons_set, transitive
    ):
        if include_selected:
            result.add(addon_name)
        if not addon:
            missing.add(addon_name)
        else:
            result.update(set(addon.manifest.depends) - set(addons_selection))
    return addon_sorter.sort(result, addons_set), missing
